# The docstring in this module is written in rst format so that it can be
# collected by sphinx and integrated into django-genes/README.rst file.

"""
   This command parses gene info file(s) and saves the corresponding
   gene objects into the database. It takes 2 required arguments and 5
   optional arguments:

   * (Required) geneinfo_file: location of gene info file;

   * (Required) taxonomy_id: taxonomy ID for organism for which genes are
     being populated;

   * (Optional) systematic_col: systematic column in gene info file.
     Default is 3;

   * (Optional) symbol_col: symbol column in gene info file. Default is 2;

   * (Optional) gi_tax_id: alternative taxonomy ID for some organisms
     (such as S. cerevisiae);

   * (Optional) alias_col: the column containing gene aliases. If a hyphen
     '-' or blank space ' ' is passed, symbol_col will be used. Default is 4.

   * (Optional) put_systematic_in_xrdb: name of cross-reference Database
     for which you want to use organism systematic IDs as CrossReference
     IDs. This is useful for Pseudomonas, for example, as systematic IDs
     are saved into PseudoCAP cross-reference database.

   The following example shows how to download a gzipped human gene
   info file from NIH FTP server, and populate the database based on
   this file.

   ::

      # Create a temporary data directory:
      mkdir data

      # Download a gzipped human gene info file into data directory:
      wget -P data/ -N \
ftp://ftp.ncbi.nih.gov/gene/DATA/GENE_INFO/Mammalia/Homo_sapiens.gene_info.gz

      # Unzip downloaded file:
      gunzip -c data/Homo_sapiens.gene_info.gz > data/Homo_sapiens.gene_info

      # Call genes_load_geneinfo to populate the database:
      python manage.py genes_load_geneinfo \
--geneinfo_file=data/Homo_sapiens.gene_info \
--taxonomy_id=9606 --systematic_col=2 --symbol_col=2
"""

import logging
import sys
from optparse import make_option

from django.core.management.base import BaseCommand
from genes.models import Gene, CrossRefDB, CrossRef
from organisms.models import Organism

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--geneinfo_file', action='store', dest='geneinfo',
                    help="gene_info file available for download from NCBI "
                    "entrez"),
        make_option('--taxonomy_id', action='store', dest='taxonomy_id',
                    help="taxonomy_id assigned by NCBI to this organism"),

        make_option('--gi_tax_id', action='store', dest='gi_tax_id',
                    help="To work with cerivisiae's tax id change, use this, "
                    "otherwise we will just use tax_id"),

        make_option('--symbol_col', action='store', dest='symbol_col',
                    help="The column containing the symbol id.", default=2),

        make_option('--systematic_col', action='store', dest='systematic_col',
                    help="The column containing the systematic id.  If this is"
                    " '-' or blank, the symbol will be used", default=3),

        make_option('--alias_col', action='store', dest='alias_col',
                    help="The column containing gene aliases.  If this is '-'"
                    "or blank, the symbol will be used", default=4),

        make_option('--put_systematic_in_xrdb', action='store',
                    dest='systematic_xrdb',
                    help='Optional: Name of Cross-Reference Database for '
                    'which you want to use organism systematic IDs as '
                    'CrossReference IDs (Used for Pseudomonas)'),
    )

    help = 'Add standards from stds_file with the associations from '\
           'assoc_file.'

    def handle(self, *args, **options):
        # Load the organism.
        tax_id = options.get('taxonomy_id')
        org = Organism.objects.get(taxonomy_id=tax_id)

        # geneinfo file information.
        geneinfo_filename = options.get('geneinfo')
        symb_col = int(options.get('symbol_col'))
        syst_col = int(options.get('systematic_col'))
        alias_col = int(options.get('alias_col'))
        systematic_xrdb = options.get('systematic_xrdb')

        # Open the geneinfo file.
        if geneinfo_filename:
            geneinfo_fh = open(geneinfo_filename)

        # yeast has a taxonomy_id that changed, in this case when we look at
        # the id from NCBI we have to use the new one.
        gi_tax_id = tax_id
        if options.get('gi_tax_id'):
            gi_tax_id = options.get('gi_tax_id')

        # Get all genes for this organism from the database.
        entrez_in_db = set(Gene.objects.filter(organism=org).values_list(
            'entrezid', flat=True))

        # Get all cross reference pairs that refer to a gene from this
        # organism.
        xr_in_db = set()
        for x in CrossRef.objects.filter(
                gene__entrezid__in=entrez_in_db).prefetch_related('crossrefdb',
                                                                  'gene'):
            xr_in_db.add((x.crossrefdb.name, x.xrid, x.gene.entrezid))

        if tax_id and geneinfo_fh:
            # Store all the genes seen thus far so we can remove obsolete
            # entries.
            entrez_seen = set()
            # Store all the crossref pairs seen thus far to avoid duplicates.
            # Cache of cross reference databases, which saves hits to DB.
            xrdb_cache = {}
            # Check to make sure the organism matched so that we don't mass-
            # delete for no reason.
            org_matches = 0
            entrez_found = 0    # Found from before.
            entrez_updated = 0  # Found from before and updated.
            entrez_created = 0  # Didn't exist, added.
            for line in geneinfo_fh:
                toks = line.strip().split('\t')
                if not (toks[0] == gi_tax_id):  # From wrong organism, skip.
                    continue

                org_matches += 1  # Count lines that came from this organism.
                # Grab requested fields from tab delimited file.
                (entrezid, standard_name, systematic_name, aliases, crossrefs,
                 description, status, chromosome) = (
                     int(toks[1]), toks[symb_col], toks[syst_col],
                     toks[alias_col], toks[5], toks[8], toks[9], toks[6])

                # This column only gets filled in for certain organisms.
                if (not systematic_name) or (systematic_name == '-'):
                    systematic_name = standard_name
                # Gene is actually mitochondrial, change symbol to avoid
                # duplicates (analogous to what GeneCards does).
                if chromosome == "MT":
                    if not systematic_name.startswith('MT'):
                        logger.debug(
                            "Renaming %s to %s, mitochondrial version",
                            systematic_name, "MT-" + systematic_name)
                        systematic_name = "MT-" + systematic_name

                alias_str = ""
                alias_num = 0
                if aliases and (aliases != '-'):
                    alias_list = [unicode(x) for x in aliases.split('|')]
                    alias_num = len(alias_list)
                    alias_str = ' '.join(alias_list)

                # Handle cross references.
                xref_tuples = []
                if crossrefs and (crossrefs != '-'):
                    xref_tuples = set()
                    if (systematic_xrdb):
                        xref_tuples.add((unicode(systematic_xrdb),
                                         unicode(systematic_name)))

                    xrefs = [unicode(x) for x in crossrefs.split('|')]
                    for x in xrefs:
                        xref_tuples.add(tuple(x.split(':')))

                xref_num = len(xref_tuples)
                # Arbitrary weight for search results.
                # The principle of weighting is that we think people are more
                # likely to want a gene that occurs in more databases or has
                # more aliases b/c it is better-known.  This helps break
                # ordering ties where gene names are identical.
                weight = 2 * xref_num + alias_num

                # We also assume that people are much more likely to want
                # protein coding genes.  In the long term we could measure
                # actual selections and estimate weight per gene.
                if status == 'protein-coding':
                    weight = weight * 2

                gene_object = None
                entrez_seen.add(entrezid)
                if entrezid in entrez_in_db:  # This existed already.
                    logger.debug("Entrez %s existed already.", entrezid)
                    entrez_found += 1
                    gene_object = Gene.objects.get(entrezid=entrezid,
                                                   organism=org)
                    changed = False
                    # The following lines update characteristics that may have
                    # changed.
                    if gene_object.systematic_name != systematic_name:
                        gene_object.systematic_name = systematic_name
                        changed = True
                    if gene_object.standard_name != standard_name:
                        gene_object.standard_name = standard_name
                        changed = True
                    if gene_object.description != description:
                        gene_object.description = description
                        changed = True
                    if gene_object.aliases != alias_str:
                        gene_object.aliases = alias_str
                        changed = True
                    if gene_object.weight != weight:
                        gene_object.weight = weight
                        changed = True
                    # If the gene was marked obsolete but occurs in the
                    # gene_info file, then it's not obsolete.
                    if gene_object.obsolete:
                        gene_object.obsolete = False
                        changed = True
                    if changed:
                        entrez_updated += 1
                        # To save time, only call save() if something has been
                        # changed.
                        gene_object.save()

                else:  # New entrezid observed.
                    logger.debug(
                        "Entrez %s did not exist and will be created.",
                        entrezid)
                    gene_object = Gene(entrezid=entrezid, organism=org,
                                       systematic_name=systematic_name,
                                       standard_name=standard_name,
                                       description=description, obsolete=False,
                                       weight=weight)
                    gene_object.save()
                    entrez_created += 1

                # Add crossreferences.
                for xref_tuple in xref_tuples:
                    try:
                        xrdb = xrdb_cache[xref_tuple[0]]
                    except KeyError:
                        try:
                            xrdb = CrossRefDB.objects.get(name=xref_tuple[0])
                        except CrossRefDB.DoesNotExist:
                            xrdb = None
                        xrdb_cache[xref_tuple[0]] = xrdb
                    if xrdb is None:  # Don't understand crossrefdb, skip.
                        logger.warning("We encountered an xrdb (%s) not in our"
                                       " database for pair %s.",
                                       xref_tuple[0], xref_tuple)
                        continue
                    logger.debug('Found crossreference pair %s.', xref_tuple)
                    # If the record doesn't exist in database, create it.
                    if not (xref_tuple[0], xref_tuple[1],
                            entrezid) in xr_in_db:
                        xr_obj = CrossRef(crossrefdb=xrdb, xrid=xref_tuple[1],
                                          gene=gene_object)
                        xr_obj.save()

            # Update "obsolete" attribute for entrez records that are in the
            # database but not in input file.
            for id in entrez_in_db:
                if id not in entrez_seen:
                    gene_object = Gene.objects.get(entrezid=id, organism=org)
                    if not gene_object.obsolete:
                        gene_object.obsolete = True
                        gene_object.save()

            logger.info("%s entrez identifiers existed in the database and "
                        "were found in the new gene_info file",
                        entrez_found)
            logger.info("%s entrez identifiers existed in the database and "
                        "were changed in the new gene_info file",
                        entrez_updated)
            logger.info("%s entrez identifiers did not exist and were created"
                        "in the new gene_info file",
                        entrez_created)
            if org_matches < 10:
                logger.error('Less than ten matches were encountered for '
                             'this organism.  Check the organism ID.')
                sys.exit(1)
        else:
            logger.error('Couldn\'t load geneinfo %s for org %s.',
                         options.get('geneinfo'),
                         tax_id,
                         exc_info=sys.exc_info(),
                         extra={'options': options})



from django.db.models.functions import Coalesce
from genes.models import Gene, CrossRef, CrossRefDB


def translate_genes(id_list=None, from_id=None, to_id=None, organism=None):
    """
    Pass a list of identifiers (id_list), the name of the database ('Entrez',
    'Symbol', 'Standard name', 'Systematic name' or a loaded crossreference
    database) that you wish to translate from, and the name of the database
    that you wish to translate to.
    """

    ids = set(id_list)
    # Initialize set of identifiers not found by this translate_genes method.
    not_found = set()
    from_ids = None  # Get the map of from_ids to the gene pks

    if organism is not None:
        gene_objects_manager = Gene.objects.filter(
            organism__scientific_name=organism)
    else:
        gene_objects_manager = Gene.objects

    if (from_id == 'Entrez'):
        int_list = []
        for x in ids:
            try:
                int_list.append(int(x))
            except(ValueError):
                not_found.add(x)
        ids = set(int_list)
        from_ids = gene_objects_manager.filter(entrezid__in=ids).values_list(
            'entrezid', 'id')
    elif (from_id == 'Systematic name'):
        from_ids = gene_objects_manager.filter(
            systematic_name__in=ids).values_list('systematic_name', 'id')
    elif (from_id == 'Standard name'):
        from_ids = gene_objects_manager.filter(
            standard_name__in=ids).values_list('standard_name', 'id')
    elif (from_id == 'Symbol'):
        # If standard_name exists, symbol will be standard_name; otherwise
        # symbol will be systematic_name.
        from_ids = gene_objects_manager.annotate(
            symbol=Coalesce('standard_name', 'systematic_name')).filter(
                symbol__in=ids).values_list('symbol', 'id')
    else:  # a crossreference db?
        xrdb = CrossRefDB.objects.get(name=from_id)
        from_ids = CrossRef.objects.filter(crossrefdb=xrdb).values_list(
            'xrid', 'gene__id')

    # Dictionary that maps from type ID passed by user to gene__id.
    from_id_map = {}
    gene_ids = []
    for item in from_ids:
        from_id_map[item[0]] = item[1]
        gene_ids.append(item[1])

    # Now let's figure out what we need to go to:
    to_ids = None
    if (to_id == 'Entrez'):
        to_ids = Gene.objects.filter(id__in=gene_ids).values_list(
            'id', 'entrezid')
    elif (to_id == 'Systematic name'):
        to_ids = Gene.objects.filter(id__in=gene_ids).values_list(
            'id', 'systematic_name')
    elif (to_id == 'Standard name'):
        to_ids = Gene.objects.filter(id__in=gene_ids).values_list(
            'id', 'standard_name')
    elif (to_id == 'Symbol'):
        # If standard_name exists, symbol will be standard_name; otherwise
        # symbol will be systematic_name.
        to_ids = Gene.objects.annotate(
            symbol=Coalesce('standard_name', 'systematic_name')).filter(
                id__in=gene_ids).values_list('id', 'symbol')

    else:  # A crossreference db?
        xrdb = CrossRefDB.objects.get(name=to_id)
        to_ids = CrossRef.objects.filter(crossrefdb=xrdb).values_list(
            'gene__id', 'xrid')

    to_id_map = {}
    for item in to_ids:
        if not item[0] in to_id_map:
            to_id_map[item[0]] = [item[1], ]
        else:
            to_id_map[item[0]].append(item[1])

    from_to = {}
    for item in ids:
        try:
            gene_id = from_id_map[item]
        except KeyError:
            not_found.add(item)
            continue
        to_id = to_id_map[gene_id]
        from_to[item] = to_id
    from_to['not_found'] = list(not_found)

    return from_to



# The docstring in this module is written in rst format so that it can be
# collected by sphinx and integrated into django-genes/README.rst file.

"""
   This command can be used to populate database with UniProtKB
   identifiers. It takes one argument:

   * uniprot_file: location of a file mapping UniProtKB IDs to Entrez
     and Ensembl IDs

   **Important:** Before calling this command, please make sure that
   both Ensembl and Entrez identifiers have been loaded into the
   database.

   After downloading the gzipped file, use ``zgrep`` command to get
   the lines we need (the original file is quite large), then run this
   command:

   ::

      wget -P data/ -N ftp://ftp.uniprot.org/pub/databases/uniprot/\
current_release/knowledgebase/idmapping/idmapping.dat.gz
      zgrep -e "GeneID" -e "Ensembl" data/idmapping.dat.gz \
> data/uniprot_entrez_ensembl.txt
      python manage.py genes_load_uniprot \
--uniprot_file=data/uniprot_entrez_ensembl.txt
"""

import logging
import sys
from optparse import make_option

from django.core.management.base import BaseCommand
from genes.models import CrossRefDB, CrossRef, Gene

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--uniprot_file', action='store', dest='uniprot',
                    help='Filtered uniprot file (i.e. with: zgrep -e "GeneID"'
                         ' -e "Ensembl" idmapping.dat.gz > '
                         'uniprot_entrez_ensembl.txt'),
    )
    help = ('Add UniProtKB cross references. *Note - Ensembl Cross-reference'
            ' database must be loaded in Tribe before running this command, '
            'as it will look for Ensembl IDs if it finds no Entrez mappings.')

    def handle(self, *args, **options):
        uniprot_file = options.get('uniprot')
        if uniprot_file:

            uniprot = CrossRefDB.objects.get(name='UniProtKB')
            ensembl = CrossRefDB.objects.get(name='Ensembl')

            entrez_set = set(Gene.objects.all().values_list('entrezid',
                                                            flat=True))
            ensembl_set = set(CrossRef.objects.filter(
                                crossrefdb=ensembl).values_list('xrid',
                                                                flat=True))
            uniprot_file = open(uniprot_file)
            uniprot_entrez = {}
            uniprot_ensembl = {}

            for line in uniprot_file:
                (uniprot_id, id_type, identifier) = line.strip().split()

                if id_type == "GeneID":
                    # 'GeneID' is a mapping for entrez
                    entrez_id = int(identifier)
                    if entrez_id in entrez_set:
                        uniprot_entrez[uniprot_id] = entrez_id

                elif id_type == "Ensembl":
                    ensembl_id = identifier
                    if ensembl_id in ensembl_set:
                        uniprot_ensembl[uniprot_id] = ensembl_id

            for uniprot_id in uniprot_entrez.keys():
                gene = Gene.objects.get(entrezid=uniprot_entrez[uniprot_id])
                try:
                    uniprot_xr = CrossRef.objects.get(crossrefdb=uniprot,
                                                      xrid=uniprot_id)
                    uniprot_xr.gene = gene
                    uniprot_xr.save()
                except CrossRef.DoesNotExist:
                    uniprot_xr = CrossRef(crossrefdb=uniprot,
                                          xrid=uniprot_id, gene=gene)
                    uniprot_xr.save()

            uniprot_set = set(CrossRef.objects.filter(
                    crossrefdb=uniprot).values_list('xrid', flat=True))

            for uniprot_id, ensembl_id in uniprot_ensembl.iteritems():
                if uniprot_id not in uniprot_set:
                    # If there is already a UniProt xref with this id in
                    # the database it means that it was already added using
                    # Entrez. We are only interested in uniprot_ids that
                    # haven't been added
                    ensembl_xr = CrossRef.objects.filter(crossrefdb=ensembl,
                                                         xrid=ensembl_id)[0]

                    gene = ensembl_xr.gene
                    uniprot_xr = CrossRef(crossrefdb=uniprot,
                                          xrid=uniprot_id, gene=gene)
                    uniprot_xr.save()

            uniprot_file.close()
        else:
            logger.error("Couldn\'t load uniprot %s", options.get('uniprot'),
                         exc_info=sys.exc_info(), extra={'options': options})



from haystack import indexes
from django.db.models import Max, Min

from genes.models import Gene

try:
    from celery_haystack.indexes import CelerySearchIndex as SearchIndex

except ImportError:
    from haystack.indexes import SearchIndex

# Had to cache this. Without the cache, rebuilding the index would crush
# the database.
cache_weights = {}


class GeneIndex(SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    organism = indexes.CharField(model_attr="organism__slug")
    obsolete = indexes.BooleanField(model_attr="obsolete")
    std_name = indexes.CharField(model_attr="standard_name", null=True)
    name_length = indexes.IntegerField()

    # Autocomplete field:
    wall_of_name_auto = indexes.EdgeNgramField(model_attr='wall_of_name')

    def prepare_name_length(self, obj):
        if obj.standard_name and not obj.standard_name.isspace():
            name = obj.standard_name
        else:
            name = obj.systematic_name
        return len(name)

    def prepare(self, obj):
        data = super(GeneIndex, self).prepare(obj)
        # Had to cache these, otherwise too much CPU would be used on
        # aggregate calls.
        try:
            (min_weight, max_weight) = cache_weights[obj.organism.slug]
        except KeyError:
            min_weight = Gene.objects.filter(organism=obj.organism).aggregate(
                Min('weight'))['weight__min']
            max_weight = Gene.objects.filter(organism=obj.organism).aggregate(
                Max('weight'))['weight__max']
            cache_weights[obj.organism.slug] = (min_weight, max_weight)

        # Boost by at most 10% for genes that are widely referred to.  This
        # helps to solve the duplicate mapping problem. See:
        # https://django-haystack.readthedocs.org/en/latest/boost.html
        # as well as management/commands/genes_load_geneinfo.py to estimate a
        # weight.
        try:
            data['boost'] = 0.1 * (obj.weight - min_weight) / (
                max_weight - min_weight) + 1
        # On the first load, the max and min weights are zero.
        except ZeroDivisionError:
            data['boost'] = 1

        return data

    def get_model(self):
        return Gene

    def index_queryset(self, using=None):
        return self.get_model().objects.all()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('organisms', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CrossRef',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('xrid', models.CharField(max_length=32, db_index=True)),
            ],
        ),
        migrations.CreateModel(
            name='CrossRefDB',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=64, db_index=True)),
                ('url', models.URLField()),
            ],
            options={
                'verbose_name': 'Cross Reference Database',
            },
        ),
        migrations.CreateModel(
            name='Gene',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('entrezid', models.IntegerField(unique=True, null=True, db_index=True)),
                ('systematic_name', models.CharField(max_length=32, db_index=True)),
                ('standard_name', models.CharField(max_length=32, null=True, db_index=True)),
                ('description', models.TextField()),
                ('aliases', models.TextField()),
                ('obsolete', models.BooleanField(default=False)),
                ('weight', models.FloatField(default=1)),
                ('organism', models.ForeignKey(to='organisms.Organism')),
            ],
        ),
        migrations.AddField(
            model_name='crossref',
            name='crossrefdb',
            field=models.ForeignKey(to='genes.CrossRefDB'),
        ),
        migrations.AddField(
            model_name='crossref',
            name='gene',
            field=models.ForeignKey(to='genes.Gene'),
        ),
    ]



import re

from django.db import models
from django.core.exceptions import FieldError
from organisms.models import Organism

# Regular expressions to be matched in the methods of 'Gene'.
nonalpha = re.compile(r'[^a-zA-Z0-9]')
num = re.compile(r'[0-9]')


class Gene(models.Model):
    """
    The class 'Gene' extends the Model class in Django. For more information,
    see: https://docs.djangoproject.com/en/dev/topics/db/models/

    Genes will be added from an online database like 'Entrez', using the Django
    management commands (see management/commands/genes_load_geneinfo.py).
    """
    entrezid = models.IntegerField(null=True, db_index=True, unique=True)

    # Used for organisms like yeast.
    systematic_name = models.CharField(max_length=32, db_index=True)

    standard_name = models.CharField(max_length=32, null=True, db_index=True)

    # Description/full name of gene.
    description = models.TextField()

    # "organism" field contains the primary key of whichever organism
    # instance the gene belongs to.  For more info, see:
    # https://docs.djangoproject.com/en/dev/ref/models/fields/#django.db.models.ForeignKey
    organism = models.ForeignKey(Organism, null=False)

    # "aliases" stores a space-separated list of aliases.
    aliases = models.TextField()

    obsolete = models.BooleanField(default=False)

    # Weight used for search results, needed now that we allow genes with
    # identical symbols.
    weight = models.FloatField(default=1)

    # __unicode__ in django, explained at:
    # https://docs.djangoproject.com/en/dev/ref/models/instances/#unicode
    def __unicode__(self):
        if self.standard_name:
            return self.standard_name
        return self.systematic_name

    def wall_of_name(self):
        '''
        Appends identifiers for the different databases (such as Entrez id's)
        and returns them. Uses the CrossRef class below.
        '''
        names = []
        if self.standard_name:
            names.append(self.standard_name)
        if self.systematic_name:
            names.append(self.systematic_name)
        names.extend([xref.xrid for xref in self.crossref_set.all()])
        for i in range(len(names)):
            names[i] = re.sub(nonalpha, '', names[i])

        names_string = ' '.join(names)
        if self.standard_name:
            names_string += ' ' + re.sub(num, '', self.standard_name)
        return names_string

    def save(self, *args, **kwargs):
        """
        Override save() method to make sure that standard_name and
        systematic_name won't be null or empty, or consist of only space
        characters (such as space, tab, new line, etc).
        """
        empty_std_name = False
        if not self.standard_name or self.standard_name.isspace():
            empty_std_name = True

        empty_sys_name = False
        if not self.systematic_name or self.systematic_name.isspace():
            empty_sys_name = True

        if empty_std_name and empty_sys_name:
            raise ValueError(
                "Both standard_name and systematic_name are empty")

        super(Gene, self).save(*args, **kwargs)  # Call the "real" save().


class CrossRefDB(models.Model):
    name = models.CharField(max_length=64, unique=True, db_index=True,
                            blank=False)
    url = models.URLField()

    def save(self, *args, **kwargs):
        """
        Extends save() method of Django models to check that the database name
        is not left blank.
        Note: 'blank=False' is only checked at a form-validation-stage. A test
        using Fixtureless that tries to randomly create a CrossRefDB with an
        empty string name would unintentionally break the test.
        """
        if self.name == '':
            raise FieldError
        else:
            return super(CrossRefDB, self).save(*args, **kwargs)

    def __unicode__(self):
        return self.name

    class Meta:
        verbose_name = "Cross Reference Database"


class CrossRef(models.Model):
    crossrefdb = models.ForeignKey(CrossRefDB, null=False)
    gene = models.ForeignKey(Gene, null=False)
    xrid = models.CharField(max_length=32, null=False, db_index=True)

    def __unicode__(self):
        return self.xrid

    # Return the url for this entry
    def _get_url(self):
        url = self.crossrefdb.url
        if "_SPEC_" in url:
            species_name = self.gene.organism.scientific_name
            url = url.replace("_SPEC_", species_name.replace(" ", "+"))
        url = url.replace("_REPL_", self.xrid)
        return url

    specific_url = property(_get_url)



import inspect
import string

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from django.core.exceptions import FieldError
from django.db import IntegrityError
from django.core.management import call_command
from fixtureless import Factory

import haystack
from haystack.query import SearchQuerySet
from tastypie.api import Api
from tastypie.test import ResourceTestCaseMixin

from organisms.models import Organism
from genes.models import Gene, CrossRef, CrossRefDB
from genes.utils import translate_genes
from genes.search_indexes import GeneIndex
from genes.app_settings import GENES_API_RESULT_LIMIT

factory = Factory()


# REQUIRES ELASTICSEARCH TO BE SETUP AS THE HAYSTACK PROVIDER.
TEST_INDEX = {
    'default': {
        'ENGINE': 'haystack.backends.elasticsearch_backend.'
                  'ElasticsearchSearchEngine',
        'URL': 'http://127.0.0.1:9200/',
        'TIMEOUT': 60 * 10,
        'INDEX_NAME': 'test_index',
    },
}

ROOT_URLCONF = getattr(settings, 'ROOT_URLCONF', None)


class GeneDBConstraintsTestCase(TestCase):
    """
    Checks that new Genes can be created in the database under different
    circumstances. Also checks that the exceptions we are looking for are
    raised when the Genes that are trying to be created do not comply with
    database constraints.
    """

    def test_std_and_sys_name_present(self):
        """
        Check that this throws no errors.
        """
        factory.create(Gene, {'standard_name': 'A1',
                              'systematic_name': 'a12'})

    def test_only_sys_name_present(self):
        """
        Check that this throws no errors.
        """
        factory.create(Gene, {'standard_name': None,
                              'systematic_name': 'b34'})

    def test_only_std_name_present(self):
        """
        Check that this throws an IntegrityError from the database when
        trying to create a Gene with a null value for systematic_name.
        """
        with self.assertRaises(IntegrityError):
            factory.create(Gene, {'standard_name': 'C5',
                                  'systematic_name': None})

    def test_both_names_absent(self):
        """
        Check that the Gene.save() method throws a ValueError when
        trying to create a Gene with a null value for standard_name AND
        systematic_name.
        """
        with self.assertRaises(ValueError):
            factory.create(Gene, {'standard_name': None,
                                  'systematic_name': None})

    def test_only_sys_name_blank_space(self):
        """
        Check that the Gene.save() method throws a ValueError if there
        is no standard_name and a systematic_name is passed but it is
        a blank string.
        """
        with self.assertRaises(ValueError):
            factory.create(Gene, {'standard_name': None,
                                  'systematic_name': ' '})

    def test_good_std_name_blank_sys_name(self):
        """
        Check that this throws no errors even though the systematic_name
        passed is a blank string (as in the previous test), since a
        non-blank standard_name is passed.
        """
        factory.create(Gene, {'standard_name': 'D7',
                              'systematic_name': ' '})

    def test_std_name_and_sys_name_both_blank_space(self):
        """
        Check that the Gene.save() method throws a ValueError if
        both the standard_name and systematic_name are passed but are
        blank strings.
        """
        with self.assertRaises(ValueError):
            factory.create(Gene, {'standard_name': '  ',
                                  'systematic_name': '  '})


class TranslateTestCase(TestCase):
    def setUp(self):
        org = factory.create(Organism)
        xrdb1 = CrossRefDB(name="ASDF", url="http://www.example.com")
        xrdb1.save()
        xrdb2 = CrossRefDB(name="XRDB2", url="http://www.example.com/2")
        xrdb2.save()

        # g1 and g2 have both standard and systematic names.
        g1 = Gene(entrezid=1, systematic_name="g1", standard_name="G1",
                  description="asdf", organism=org, aliases="gee1 GEE1")
        g1.save()
        g2 = Gene(entrezid=2, systematic_name="g2", standard_name="G2",
                  description="asdf", organism=org, aliases="gee2 GEE2")
        g2.save()

        xref1 = CrossRef(crossrefdb=xrdb1, gene=g1, xrid="XRID1")
        xref1.save()
        xref2 = CrossRef(crossrefdb=xrdb2, gene=g2, xrid="XRID1")
        xref2.save()
        xref3 = CrossRef(crossrefdb=xrdb1, gene=g1, xrid="XRRID1")
        xref3.save()
        xref4 = CrossRef(crossrefdb=xrdb1, gene=g2, xrid="XRID2")
        xref4.save()

        org2 = Organism(taxonomy_id=1234, common_name="Computer mouse",
                        scientific_name="Mus computurus",
                        slug="mus-computurus")
        org2.save()
        org3 = Organism(taxonomy_id=4321, common_name="Computer screen",
                        scientific_name="Monitorus computurus",
                        slug="monitorus-computurus")
        org3.save()

        # Make systematic and standard name the same for the following genes,
        # but make organisms different. Skip entrezid 3 since that is used by
        # other tests.
        g4 = Gene(entrezid=4, systematic_name="acdc", standard_name="ACDC",
                  description="asdf", organism=org2, aliases="gee4 GEE4")
        g4.save()
        g5 = Gene(entrezid=5, systematic_name="acdc", standard_name="ACDC",
                  description="asdf", organism=org3, aliases="gee5 GEE5")
        g5.save()

        # g101 has standard name, but no systematic name.
        g101 = Gene(entrezid=101, standard_name="std_101", organism=org2)
        g101.save()

        # g102 has systematic name, but no standard name.
        g102 = Gene(entrezid=102, systematic_name="sys_102", organism=org2)
        g102.save()

    def test_translate_symbol_entrez_diff_organisms(self):
        """
        translate_genes() should be able to differentiate between different
        organism genes when passed identical symbols.
        """
        # This test also confirmed that when both standard name and systematic
        # name are available, the sysmbol will be standard name.
        translation = translate_genes(id_list=['ACDC'],
                                      from_id="Symbol", to_id="Entrez",
                                      organism="Mus computurus")
        self.assertEqual(translation, {'ACDC': [4], 'not_found': []})

    def test_translate_symbol_entrez_diff_organisms2(self):
        """
        Same as previous test, but uses the other organism as input.
        """
        translation = translate_genes(id_list=['ACDC'],
                                      from_id="Symbol", to_id="Entrez",
                                      organism="Monitorus computurus")
        self.assertEqual(translation, {'ACDC': [5], 'not_found': []})

    def test_translate_entrez_entrez(self):
        """
        Test translation from entrez to entrez.
        """
        translation = translate_genes(id_list=[1, 2],
                                      from_id="Entrez", to_id="Entrez")
        self.assertEqual(translation, {1: [1, ], 2: [2, ], 'not_found': []})

    def test_translate_entrez_standard_name(self):
        """
        Test translation from entrez to standard names.
        """
        translation = translate_genes(id_list=[1, 2],
                                      from_id="Entrez",
                                      to_id="Standard name")
        self.assertEqual(translation,
                         {1: ['G1', ], 2: ['G2', ], 'not_found': []})

    def test_translate_entrez_systematic_name(self):
        """
        Test translation from entrez to systematic names.
        """
        translation = translate_genes(id_list=[1, 2],
                                      from_id="Entrez",
                                      to_id="Systematic name")
        self.assertEqual(translation,
                         {1: ['g1', ], 2: ['g2', ], 'not_found': []})

    def test_translate_entrez_xrdb(self):
        """
        Test translation from entrez to ASDF.
        """
        translation = translate_genes(id_list=[1, 2],
                                      from_id="Entrez", to_id="ASDF")
        self.assertEqual(translation, {1: ['XRID1', 'XRRID1', ],
                                       2: ['XRID2', ], 'not_found': []})

    def test_translate_xrdb_entrez(self):
        """
        Test translation from ASDF to entrez.
        """
        translation = translate_genes(id_list=['XRID1', 'XRRID1', 'XRID2'],
                                      from_id="ASDF", to_id="Entrez")
        self.assertEqual(translation, {'XRID1': [1, ], 'XRRID1': [1, ],
                                       'XRID2': [2, ], 'not_found': []})

    def test_translate_entrez_entrez_missing(self):
        """
        Test translation from entrez to entrez with a missing value.
        """
        translation = translate_genes(id_list=[1, 2, 3],
                                      from_id="Entrez", to_id="Entrez")
        self.assertEqual(translation, {1: [1, ], 2: [2, ], 'not_found': [3]})

    def test_translate_entrez_standard_name_missing(self):
        """
        Test translation from entrez to standard names with a missing value.
        """
        translation = translate_genes(id_list=[1, 2, 3],
                                      from_id="Entrez", to_id="Standard name")
        self.assertEqual(translation,
                         {1: ['G1', ], 2: ['G2', ], 'not_found': [3]})

    def test_translate_symbol_entrez(self):
        """
        Test translation from symbol to entrez when either standard name or
        systematic name is null.
        """
        # Test the gene that has standard name.
        translation = translate_genes(id_list=['std_101'],
                                      from_id="Symbol", to_id="Entrez",
                                      organism="Mus computurus")
        self.assertEqual(translation, {'std_101': [101], 'not_found': []})
        # Test the gene that does NOT have standard name.
        translation = translate_genes(id_list=['sys_102'],
                                      from_id="Symbol", to_id="Entrez",
                                      organism="Mus computurus")
        self.assertEqual(translation, {'sys_102': [102], 'not_found': []})

    def test_translate_entrez_symbol(self):
        """
        Test translation from entrez to symbol when either standard name or
        systematic name is null.
        """
        # Test the gene that has standard name.
        translation = translate_genes(id_list=[101],
                                      from_id="Entrez", to_id="Symbol",
                                      organism="Mus computurus")
        self.assertEqual(translation, {101: ['std_101'], 'not_found': []})
        # Test the gene that does NOT have standard name.
        translation = translate_genes(id_list=[102],
                                      from_id="Entrez", to_id="Symbol",
                                      organism="Mus computurus")
        self.assertEqual(translation, {102: ['sys_102'], 'not_found': []})

    def test_empty_standard_and_systematic_names(self):
        """
        Test that a ValueError exception will be raised when we try to create a
        gene whose standard and systematic names are both empty or null, or
        ONLY consist of space characters (such as space, tab, new line, etc).
        """
        org = factory.create(Organism)

        # Neither standard_name nor systematic_name is set explicitly.
        unnamed_gene = Gene(entrezid=999, organism=org)
        self.assertRaises(ValueError, unnamed_gene.save)

        # standard_name consists of only space characters.
        # systematic_name is u'' here, because it is not set explicitly, and
        # by default "null=False" for this field in the model.
        unnamed_gene = Gene(entrezid=999, standard_name="\t  \n", organism=org)
        self.assertRaises(ValueError, unnamed_gene.save)

        # Both standard_name and systematic_name are empty strings.
        unnamed_gene = Gene(entrezid=999, standard_name="", systematic_name="",
                            organism=org)
        self.assertRaises(ValueError, unnamed_gene.save)

        # Both standard_name and systematic_name consist of space characters
        # only.
        unnamed_gene = Gene(entrezid=999, standard_name="  ",
                            systematic_name="\t  \n ", organism=org)
        self.assertRaises(ValueError, unnamed_gene.save)

    def tearDown(self):
        Organism.objects.all().delete()    # Remove Organism objects.
        Gene.objects.all().delete()        # Remove Gene objects.
        CrossRef.objects.all().delete()    # Remove CrossRef objects.
        CrossRefDB.objects.all().delete()  # Remove CrossRefDB objects.


class PrepareNameLengthTestCase(TestCase):
    """
    This TestCase prepares the prepare_name_length() method in
    search_indexes.GeneIndex.
    """
    def setUp(self):

        self.g1 = factory.create(Gene, {'standard_name': 'A1',
                                        'systematic_name': 'a12'})
        self.g2 = factory.create(Gene, {'standard_name': None,
                                        'systematic_name': 'b12'})

        self.gene_index = GeneIndex()

    def test_std_and_sys_name_present(self):
        """
        Test that name_length is 2, the length of 'A1'.
        """
        name_length = self.gene_index.prepare_name_length(self.g1)
        self.assertEqual(name_length, 2)

    def test_only_sys_name_present(self):
        """
        Test that name_length is 3, the length of 'b12'.
        """
        name_length = self.gene_index.prepare_name_length(self.g2)
        self.assertEqual(name_length, 3)


# We use @override_settings here so that the tests use the TEST_INDEX
# when building/rebuilding the search indexes, and not our real Database
# search indexes.
@override_settings(HAYSTACK_CONNECTIONS=TEST_INDEX)
class BuildingGeneIndexTestCase(TestCase):
    """
    This TestCase tests the ability to build search indexes under certain
    corner cases.
    """

    def setUp(self):
        haystack.connections.reload('default')

        # As per this documented issue in Haystack,
        # https://github.com/django-haystack/django-haystack/issues/704
        # we need to call 'rebuild_index' at the beginning to get
        # consistency of data and structure. Otherwise, the
        # 'test_std_and_sys_name_present' *sometimes* yielded a None
        # result
        call_command('rebuild_index', interactive=False, verbosity=0)

    def test_factory_gene_creation(self):
        """
        Create a gene using the factory, without any specified fields.

        Call command to build search index and then try to find the gene
        using this search index.
        """
        gene = factory.create(Gene)

        call_command('update_index', interactive=False, verbosity=0)
        sqs = SearchQuerySet().models(Gene)
        sqs = sqs.filter(content=gene.systematic_name).load_all()
        self.assertEqual(sqs[0].object, gene)

    def test_std_and_sys_name_present(self):
        """
        Create a gene using the factory, but specify both standard_
        and systematic_ names.

        Call command to build search index and then try to find the gene
        using this search index.
        """
        gene = factory.create(Gene, {'standard_name': 'A1',
                                     'systematic_name': 'a12'})

        call_command('update_index', interactive=False, verbosity=0)
        sqs = SearchQuerySet().models(Gene)
        sqs = sqs.filter(content=gene.systematic_name).load_all()
        self.assertEqual(sqs[0].object, gene)

    def test_only_sys_name_present(self):
        """
        Create a gene using the factory, specify systematic_ name and
        make standard_name explicitly None.

        Call command to build search index and then try to find the gene
        using this search index.

        """
        gene = factory.create(Gene, {'standard_name': None,
                                     'systematic_name': 'b34'})

        call_command('update_index', interactive=False, verbosity=0)
        sqs = SearchQuerySet().models(Gene)
        sqs = sqs.filter(content=gene.systematic_name).load_all()
        self.assertEqual(sqs[0].object, gene)

    def test_no_description(self):
        """
        Create a gene using the factory, specify description to be an
        empty string.

        Call command to build search index and then try to find the gene
        using this search index.

        """
        gene = factory.create(Gene, {'description': ''})

        call_command('update_index', interactive=False, verbosity=0)
        sqs = SearchQuerySet().models(Gene)
        sqs = sqs.filter(content=gene.systematic_name).load_all()
        self.assertEqual(sqs[0].object, gene)

    def tearDown(self):
        call_command('clear_index', interactive=False, verbosity=0)


# We use @override_settings here so that the tests use the TEST_INDEX
# when building/rebuilding the search indexes, and not our real Database
# search indexes.
@override_settings(HAYSTACK_CONNECTIONS=TEST_INDEX)
class APIResourceTestCase(ResourceTestCaseMixin, TestCase):
    """
    Test API endpoints for retrieving and searching gene data, using both
    GET and POST requests.
    """

    def get_api_name(self):
        """
        Utility function to get the name of the tastypie REST API in
        whatever Django project is using django-genes.
        """
        if not ROOT_URLCONF:
            return None

        proj_urls = __import__(ROOT_URLCONF)
        url_members = inspect.getmembers(proj_urls.urls)

        api_name = None
        for k, v in url_members:
            if isinstance(v, Api):
                api_name = v.api_name
        return api_name

    def create_many_genes(self, organism, num_genes):
        """
        Helper function to generate a large number of genes
        """
        # Create genes:
        for i in range(num_genes):
            Gene.objects.create(entrezid=(i + 1),
                                systematic_name="sys_name #" + str(i + 1),
                                standard_name="std_name #" + str(i + 1),
                                organism=organism)

    def setUp(self):
        haystack.connections.reload('default')

        # This line is important to set up the test case!
        super(APIResourceTestCase, self).setUp()

        self.gene1 = factory.create(Gene, {'standard_name': 'A1',
                                           'systematic_name': 'a12'})
        self.gene2 = factory.create(Gene, {'standard_name': None,
                                           'systematic_name': 'b34'})

        standard_name_prefix = 'ans'
        factory.create(Gene, {'standard_name': standard_name_prefix})

        # Create 26 more gene names that start with 'ans' and then have
        # an uppercase letter appended to it.
        for letter in string.ascii_uppercase:
            factory.create(
                Gene,
                {'standard_name': standard_name_prefix + letter}
            )

        call_command('rebuild_index', interactive=False, verbosity=0)

    def test_gene_get_search(self):
        """
        Tests API gene search when searching with a GET request
        """

        api_name = self.get_api_name()
        response = self.api_client.get(
            '/api/{}/gene/search/'.format(api_name),
            data={'query': self.gene1.standard_name}
        )

        self.assertValidJSONResponse(response)

        found_results = self.deserialize(response)[0]['found']
        best_gene_result = found_results[0]

        self.assertEqual(best_gene_result['standard_name'],
                         self.gene1.standard_name)
        self.assertEqual(best_gene_result['systematic_name'],
                         self.gene1.systematic_name)

    def test_gene_post_search(self):
        """
        Tests API gene search when searching with a POST request
        """
        api_name = self.get_api_name()

        response = self.api_client.post(
            '/api/{}/gene/search/'.format(api_name),
            data={'query': self.gene2.systematic_name}
        )

        self.assertValidJSONResponse(response)

        found_results = self.deserialize(response)[0]['found']
        best_gene_result = found_results[0]

        self.assertEqual(best_gene_result['standard_name'],
                         self.gene2.standard_name)
        self.assertEqual(best_gene_result['systematic_name'],
                         self.gene2.systematic_name)

    def test_gene_list_endpt_large_post(self):
        """
        Test that we can do a big POST request to get information back
        for a lot of genes (more than are allowed through the ~4k
        character limit for GET).

        We will set the gene_num to 1100 because

        9 * 1 = 9 chars of single-digit IDs (1-9)
        90 * 2 = 180 chars of double-digit IDs (10-99)
        900 * 3 = 2700 chars of triple-digit IDs (100-999)
        101 * 4 = 404 chars of four-digit IDs (1000-1100)
        (1100 - 1) * 1 chars of delimiters (',')

        TOTAL = 9 + 180 + 2700 + 404 + 1099 = 4392 chars.

        This is based on the
        APIResourceTestCase.test_expressionvalue_big_post() test in
        https://github.com/greenelab/adage-server/blob/master/adage/analyze/tests.py
        """
        organism = factory.create(Organism)
        gene_num = 1100

        self.create_many_genes(organism, gene_num)
        gene_ids = ",".join([str(g.id) for g
                             in Gene.objects.filter(organism=organism)])

        api_name = self.get_api_name()
        resp = self.client.post('/api/{}/gene/'.format(api_name),
                                data={'pk__in': gene_ids})

        self.assertValidJSONResponse(resp)
        self.assertEqual(
            self.deserialize(resp)['meta']['total_count'],
            gene_num
        )

    def test_gene_autocomplete_search(self):
        """
        Tests API gene autocomplete search. In the setUp method, we
        created 27 genes that start with 'ans', but this should only
        return 15 results, or however many were set in the
        GENES_API_RESULT_LIMIT setting.
        """

        api_name = self.get_api_name()

        response = self.api_client.get(
            '/api/{}/gene/autocomplete/'.format(api_name),
            data={'query': 'ans'}
        )
        self.assertValidJSONResponse(response)
        found_results = self.deserialize(response)['results']
        self.assertEqual(len(found_results), GENES_API_RESULT_LIMIT)

    def tearDown(self):
        call_command('clear_index', interactive=False, verbosity=0)


class CrossRefDBTestCase(TestCase):
    def test_saving_xrdb(self):
        """
        Test that this simple CrossRefDB creation raises no errors.
        """
        factory.create(CrossRefDB, {"name": "XRDB1"})

    def test_saving_xrdb_no_name(self):
        """
        Check that CrossRefDBs in database are required to have a non-null
        name - if they do, raise IntegrityError.
        """
        with self.assertRaises(IntegrityError):
            factory.create(CrossRefDB, {"name": None})

    def test_saving_xrdb_blank_name(self):
        """
        Check that CrossRefDBs in database are required to have a name that
        is not an empty string - if they do, raise FieldError.
        """
        with self.assertRaises(FieldError):
            factory.create(CrossRefDB, {"name": ""})


class LoadCrossRefsTestCase(TestCase):

    def setUp(self):
        xrdb1 = CrossRefDB.objects.create(
            name="Ensembl", url="http://www.ensembl.org/Gene/Summary?g=_REPL_")

        CrossRefDB.objects.create(
            name="UniProtKB", url="http://www.uniprot.org/uniprot/_REPL_")

        g1 = factory.create(Gene, {'entrezid': 50810})
        g2 = factory.create(Gene)
        g3 = factory.create(Gene)
        g4 = factory.create(Gene)

        factory.create(CrossRef, {'crossrefdb': xrdb1, 'gene': g1,
                                  'xrid': 'ENSG00000166503'})
        factory.create(CrossRef, {'crossrefdb': xrdb1, 'gene': g2,
                                  'xrid': 'ENSG00000214575'})
        factory.create(CrossRef, {'crossrefdb': xrdb1, 'gene': g3,
                                  'xrid': 'ENSG00000170312'})
        factory.create(CrossRef, {'crossrefdb': xrdb1, 'gene': g4,
                                  'xrid': 'ENSG00000172053'})

    def test_load_uniprot_mgmt_command(self):
        """
        Check that genes_load_uniprot management command loads UniProtKB
        identifiers (using Entrez and Ensembl) and that it also saves those
        relationships in the database.
        """
        call_command('genes_load_uniprot',
                     uniprot='genes/test_files/test_uniprot_entrez_ensembl.txt')

        uniprot1 = CrossRef.objects.get(xrid='A0A024R216')
        self.assertEqual(uniprot1.gene.entrezid, 50810)
        e1 = CrossRef.objects.filter(
            crossrefdb__name='Ensembl').get(gene=uniprot1.gene)
        self.assertEqual(e1.xrid, 'ENSG00000166503')

        uniprot2 = CrossRef.objects.get(xrid='A0A024R214')
        e2 = CrossRef.objects.filter(
            crossrefdb__name='Ensembl').get(gene=uniprot2.gene)
        self.assertEqual(e2.xrid, 'ENSG00000214575')

        uniprot3 = CrossRef.objects.get(xrid='A0A024QZP7')
        e3 = CrossRef.objects.filter(
            crossrefdb__name='Ensembl').get(gene=uniprot3.gene)
        self.assertEqual(e3.xrid, 'ENSG00000170312')

        uniprot4 = CrossRef.objects.get(xrid='A0A0U1RQX9')
        e4 = CrossRef.objects.filter(
            crossrefdb__name='Ensembl').get(gene=uniprot4.gene)
        self.assertEqual(e4.xrid, 'ENSG00000172053')



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-genes',
    version='0.15',
    packages=find_packages(),
    include_package_data=True,
    license='LICENSE.txt',
    description='A simple Django app to represent genes.',
    long_description=README,
    url='https://github.com/greenelab/django-genes',
    author='Greene Lab',
    author_email='team@greenelab.com',
    install_requires=[
        'django>=1.8',
        'django-organisms>=0.5',
        'django-haystack',
        'django-fixtureless',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



# The docstring in this module is written in rst format so that it can be
# collected by sphinx and integrated into django-genes/README.rst file.

"""
   This command can be used to populate database with WormBase
   identifiers. It takes 3 arguments:

   * (Required) wb_url: URL of wormbase xrefs file;

   * (Required) taxonomy_id: taxonomy ID assigned to this organism by
     NCBI;

   * (Optional) db_name: the name of the cross-reference database,
     default is 'WormBase'.

   As is expected, the WormBase cross-reference database should be
   populated using the ``genes_add_xrdb`` command (see command #1)
   before this command to populate the WormBase identifiers.
   Here is an example:

   ::

      # Find latest version of WormBase here:
      # http://www.wormbase.org/about/release_schedule#102--10-1
      python manage.py genes_load_wb --wb_url=ftp://ftp.wormbase.org/pub/\
wormbase/releases/WS243/species/c_elegans/PRJNA13758/\
c_elegans.PRJNA13758.WS243.xrefs.txt.gz --taxonomy_id=6239
"""

import logging
import urllib2
import gzip
from optparse import make_option
from StringIO import StringIO

from django.core.management.base import BaseCommand
from genes.models import Gene, CrossRefDB, CrossRef

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--wb_url', action='store', dest='wburl',
                    help="URL of wormbase xrefs file."),
        make_option('--db_name', action='store', dest='dbname',
                    help="Name of the database- defaults to 'WormBase'.",
                    default="WormBase"),
        make_option('--taxonomy_id', action='store', dest='taxonomy_id',
                    help="taxonomy_id assigned by NCBI to this organism"),
    )

    help = 'Add wormbase identifiers to database.'

    def handle(self, *args, **options):
        database = CrossRefDB.objects.get(name=options.get('dbname'))
        wb_url = options.get('wburl')

        xrefs_gzip_fh = gzip.GzipFile(fileobj=StringIO(
            urllib2.urlopen(wb_url, timeout=5).read()))

        for line in xrefs_gzip_fh:
            toks = line.strip().split('\t')
            systematic = 'CELE_' + toks[0]
            wbid = toks[1]
            try:
                gene = Gene.objects.get(systematic_name=systematic)
            except Gene.DoesNotExist:
                logger.info("Unable to find gene %s.", systematic)
                continue
            wb = None
            try:
                wb = CrossRef.objects.get(xrid=wbid, crossrefdb=database)
            except CrossRef.DoesNotExist:
                wb = CrossRef(xrid=wbid, crossrefdb=database)
            wb.gene = gene
            wb.save()

        xrefs_gzip_fh.close()



# The docstring in this module is written in rst format so that it can be
# collected by sphinx and integrated into django-genes/README.rst file.

"""
   This command adds cross-reference databases for genes. It **must**
   be called for every new cross-reference database to populate the
   gene and cross-reference objects in the database. It requires 2
   arguments:

   * name: the name of the database

   * URL: the URL for that database, with the string '_REPL_' added at the
     end of the URL

   For example, this command adds Ensembl as a cross-reference
   database:

   ::

      python manage.py genes_add_xrdb --name=Ensembl \
--URL=http://www.ensembl.org/Gene/Summary?g=_REPL_

   And this command adds MIM as a cross-reference database:

   ::

      python manage.py genes_add_xrdb --name=MIM \
--URL=http://www.ncbi.nlm.nih.gov/omim/_REPL_
"""

import logging
from optparse import make_option

from django.core.management.base import BaseCommand
from genes.models import CrossRefDB

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--name', action='store', dest='name'),
        make_option('--URL', action='store', dest='url'),
    )

    help = 'Add a cross reference database if one with the provided name '\
           'does not exist, or update the URL if it does.'

    def handle(self, *args, **options):
        name = options.get('name', None)
        url = options.get('url', None)
        try:
            xrdb = CrossRefDB.objects.get(name=name)
            if xrdb.url != url:
                xrdb.url = url
                xrdb.save()
        except CrossRefDB.DoesNotExist:
            xrdb = CrossRefDB(name=name, url=url)
            xrdb.save()



#!/usr/bin/env python

# The docstring in this module is written in rst format so that it can be
# collected by sphinx and integrated into django-genes/README.rst file.

"""
   This management command will read an input gene history file and
   find all genes whose tax_id match input taxonomy ID. If the gene
   already exists in the database, the Gene record in database will be
   set as obsolete; if not, a new obsolete Gene record will be created
   in the database.

   The command accepts 2 required arguments and 3 optional arguments:

   * (Required) gene_history_file: Input gene history file. A gzipped
     example file can be found at:
     ftp://ftp.ncbi.nih.gov/gene/DATA/gene_history.gz

   * (Required) tax_id: Taxonomy ID assigned by NCBI to a certain
     organism. Genes of the other organisms in input file will be
     skipped.

   * (Optional) tax_id_col: column number of tax_id in input file.
     Default is 1.

   * (Optional) discontinued_id_col: column number of discontinued
     GeneID in input file. Default is 3.

   * (Optional) discontinued_symbol_col: column number of gene's
     discontinued symbol in input file. Default is 4.

   Note that column numbers in the last three arguments all start from
   1, **not** 0.

   For example, to add obsolete genes whose tax_id is 208964 in the
   file "gene_history", we will use the command like this:

   ::

      # Download file into your data directory:
      cd /data_dir; wget ftp://ftp.ncbi.nih.gov/gene/DATA/gene_history.gz

      # Unzip the downloaded file into "gene_history"
      gunzip gene_history.gz

      # Run management command:
      python manage.py genes_load_gene_history /data_dir/gene_history 208964 \
--tax_id_col=1 --discontinued_id_col=3 --discontinued_symbol_col=4

   (Here ``--tax_id_col=1 --discontinued_id_col=3
   --discontinued_symbol_col=4`` are optional because they are using
   default values.)
"""


from django.core.management.base import BaseCommand, CommandError
from organisms.models import Organism
from genes.models import Gene


class Command(BaseCommand):
    help = ('Read input gene_history file and set or create obsolete Gene '
            'records in the database.')

    def add_arguments(self, parser):
        parser.add_argument('gene_history_file', type=file)
        parser.add_argument('tax_id', type=str)
        parser.add_argument('--tax_id_col', type=int, default=1,
                            dest='tax_id_col',
                            help='column number of tax_id')
        parser.add_argument('--discontinued_id_col', type=int, default=3,
                            dest='id_col',
                            help='column number of Discontinued_GeneID')
        parser.add_argument('--discontinued_symbol_col', type=int, default=4,
                            dest='symbol_col',
                            help='column number of Discontinued_Symbol')

    def handle(self, *args, **options):
        try:
            import_gene_history(options['gene_history_file'],
                                options['tax_id'],
                                options['tax_id_col'] - 1,
                                options['id_col'] - 1,
                                options['symbol_col'] - 1)
            self.stdout.write(self.style.NOTICE(
                'Gene history data import succeeded'))
        except Exception as e:
            raise CommandError(
                'Data import encountered an error: import_gene_history throws '
                'an exception:\n%s' % e)


def chk_col_numbers(line_num, num_cols, tax_id_col, id_col, symbol_col):
    """
    Check that none of the input column numbers is out of range.
    (Instead of defining this function, we could depend on Python's built-in
    IndexError exception for this issue, but the IndexError exception wouldn't
    include line number information, which is helpful for users to find exactly
    which line is the culprit.)
    """

    bad_col = ''
    if tax_id_col >= num_cols:
        bad_col = 'tax_id_col'
    elif id_col >= num_cols:
        bad_col = 'discontinued_id_col'
    elif symbol_col >= num_cols:
        bad_col = 'discontinued_symbol_col'

    if bad_col:
        raise Exception(
            'Input file line #%d: column number of %s is out of range' %
            (line_num, bad_col))


def import_gene_history(file_handle, tax_id, tax_id_col, id_col, symbol_col):
    """
    Read input gene history file into the database.
    Note that the arguments tax_id_col, id_col and symbol_col have been
    converted into 0-based column indexes.
    """

    # Make sure that tax_id is not "" or "  "
    if not tax_id or tax_id.isspace():
        raise Exception("Input tax_id is blank")

    # Make sure that tax_id exists in Organism table in the database.
    try:
        organism = Organism.objects.get(taxonomy_id=tax_id)
    except Organism.DoesNotExist:
        raise Exception('Input tax_id %s does NOT exist in Organism table. '
                        'Please add it into Organism table first.' % tax_id)

    if tax_id_col < 0 or id_col < 0 or symbol_col < 0:
        raise Exception(
            'tax_id_col, id_col and symbol_col must be positive integers')

    for line_index, line in enumerate(file_handle):
        if line.startswith('#'):  # Skip comment lines.
            continue

        fields = line.rstrip().split('\t')
        # Check input column numbers.
        chk_col_numbers(line_index + 1, len(fields), tax_id_col, id_col,
                        symbol_col)

        # Skip lines whose tax_id's do not match input tax_id.
        if tax_id != fields[tax_id_col]:
            continue

        entrez_id = fields[id_col]
        # If the gene already exists in database, set its "obsolete" attribute
        # to True; otherwise create a new obsolete Gene record in database.
        try:
            gene = Gene.objects.get(entrezid=entrez_id)
            if not gene.obsolete:
                gene.obsolete = True
                gene.save()
        except Gene.DoesNotExist:
            Gene.objects.create(entrezid=entrez_id, organism=organism,
                                systematic_name=fields[symbol_col],
                                obsolete=True)



import json

from django.conf.urls import url
from haystack.query import SearchQuerySet

from genes.models import Gene, CrossRefDB, CrossRef
from genes.utils import translate_genes
from genes.app_settings import GENES_API_RESULT_LIMIT
from organisms.api import OrganismResource

# Import and set logger
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

try:
    from tastypie import fields
    from tastypie.resources import (ModelResource, ALL, ALL_WITH_RELATIONS,
                                    convert_post_to_VERB)
    from tastypie.utils import trailing_slash

except ImportError:
    logger.info('Not using django-tastypie in genes/api.py file')
    quit()


class GeneResource(ModelResource):
    entrezid = fields.IntegerField(attribute='entrezid')
    pk = fields.IntegerField(attribute='id')
    xrids = fields.ToManyField('genes.api.CrossRefResource',
                               'crossref_set', related_name='gene',
                               full=True)

    class Meta:
        queryset = Gene.objects.all()
        resource_name = 'gene'
        filtering = {'entrezid': ALL, 'pk': ALL, 'symbol': ALL}
        list_allowed_methods = ['get', 'post']
        detail_allowed_methods = ['get']

    def prepend_urls(self):
        return [
            url(r"^(?P<resource_name>%s)/search%s$" %
                (self._meta.resource_name, trailing_slash()),
                self.wrap_view('get_search'), name="api_get_search"),
            url(r"^(?P<resource_name>%s)/autocomplete%s$" %
                (self._meta.resource_name, trailing_slash()),
                self.wrap_view('autocomplete'),
                name="api_autocomplete"),
            url(r"^(?P<resource_name>%s)/xrid_translate%s$" %
                (self._meta.resource_name, trailing_slash()),
                self.wrap_view('translate_gene_ids'),
                name="api_translate_gene_ids"),
        ]

    def get_search(self, request, **kwargs):
        # This code uses Haystack to search for genes
        self.method_check(request, allowed=['get', 'post'])
        self.throttle_check(request)

        if request.method == 'GET':
            data_dict = request.GET

        elif request.method == 'POST':
            data_dict = json.loads(request.body)

        limit = data_dict.get('limit')
        search_string = data_dict.get('query')
        organism_uri = data_dict.get('organism')

        if limit:
            try:
                # Make sure the input is already an integer
                # or can be coerced into one.
                limit = int(limit)
            except ValueError:
                logger.error("'limit' parameter passed was specified "
                             "incorrectly.")
                # Limit was specified incorrectly, so default to
                # GENES_API_RESULT_LIMIT setting.
                limit = GENES_API_RESULT_LIMIT

        else:
            limit = GENES_API_RESULT_LIMIT

        organism = None
        if organism_uri:
            organism = OrganismResource().get_via_uri(organism_uri,
                                                      request)
        if search_string:
            search_toks = set(search_string.split())
        else:
            search_toks = []

        genes = []

        for search in search_toks:
            item = {}
            item['search'] = search
            sqs = SearchQuerySet().models(Gene)
            if organism is not None:
                sqs = sqs.filter(organism=organism)
            sqs = sqs.filter(content=search)
            sqs = sqs.load_all()[:limit]
            objects = []

            for result in sqs:
                bundle = self.build_bundle(obj=result.object,
                                           request=request)
                bundle = self.full_dehydrate(bundle)
                objects.append(bundle)

            item['found'] = objects
            genes.append(item)

        self.log_throttled_access(request)
        return self.create_response(request, genes)

    def autocomplete(self, request, **kwargs):
        # Note: If the parameter in question (e.g. query, limit, etc.)
        # is not sent as part of the request, the request.GET.get()
        # method will set it to None unless a default is specified.
        query = request.GET.get('query', '')
        organism_uri = request.GET.get('organism')
        limit = request.GET.get('limit')

        if limit:
            try:
                # Make sure the input is already an integer
                # or can be coerced into one.
                limit = int(limit)
            except ValueError:
                logger.error("'limit' parameter passed was specified "
                             "incorrectly.")
                # Limit was specified incorrectly, so default to
                # GENES_API_RESULT_LIMIT setting.
                limit = GENES_API_RESULT_LIMIT
        else:
            limit = GENES_API_RESULT_LIMIT

        # We want to sort results by three fields: First by search score,
        # then by name length (standard_name if present, if not then
        # systematic_name), and finally by this name's alphabetical order.
        # We sort by gene name length because a long gene name and
        # a short gene name can have the same score if they contain the n-gram.
        # A user can always type more to get the long one, but they can't type
        # less to get the short one. By returning the shortest first, we make
        # sure that this choice is always available to the user, even if a
        # limit is applied.
        #
        # *Note: We use SearchQuerySet's order_by() function to sort by these
        # three fields, but this current implementation is Elasticsearch
        # specific. (Elasticsearch uses '_score' while it is simply 'score' in
        # Solr). At some point we might want to expand this to be compatible
        # with other search engines, like Solr, Xapian and Whoosh.
        # See Haystack issue here:
        # https://github.com/django-haystack/django-haystack/issues/1431
        sqs = SearchQuerySet().models(Gene).autocomplete(
            wall_of_name_auto=query).order_by('-_score', 'name_length',
                                              'wall_of_name_auto')

        if organism_uri:
            organism = OrganismResource().get_via_uri(organism_uri,
                                                      request)
            sqs = sqs.filter(organism=organism)

        suggestions = []

        # Get slice of sqs for 'limit' specified (or use full sqs if no limit
        # was specified).
        for result in sqs[:limit]:
            gene = result.object
            suggestions.append({'id': gene.id, 'score': result.score,
                                'entrezid': gene.entrezid,
                                'standard_name': gene.standard_name,
                                'systematic_name': gene.systematic_name,
                                'description': gene.description})

        # Return a JSON object instead of an array, as returning an array
        # could make the information vulnerable to an XSS attack.
        response = {'results': suggestions}
        return self.create_response(request, response)

    def translate_gene_ids(self, request, **kwargs):
        self.method_check(request, allowed=['post'])
        self.throttle_check(request)

        # Extract relevant parameters
        gene_list = request.POST.getlist('gene_list')
        from_id = request.POST.get('from_id')
        to_id = request.POST.get('to_id')
        organism = request.POST.get('organism')

        # Call the translate_genes method in genes.utils to
        # do the actual translating.
        return_ids_dict = translate_genes(id_list=gene_list,
                                          from_id=from_id, to_id=to_id,
                                          organism=organism)

        return self.create_response(request, return_ids_dict)

    def post_list(self, request, **kwargs):
        """
        (Copied from implementation in
        https://github.com/greenelab/adage-server/blob/master/adage/analyze/api.py)

        Handle an incoming POST as a GET to work around URI length limitations
        """
        # The convert_post_to_VERB() technique is borrowed from
        # resources.py in tastypie source. This helps us to convert the POST
        # to a GET in the proper way internally.
        request.method = 'GET'  # override the incoming POST
        dispatch_request = convert_post_to_VERB(request, 'GET')
        return self.dispatch('list', dispatch_request, **kwargs)


class CrossRefDBResource(ModelResource):
    name = fields.CharField(attribute='name')
    url = fields.CharField(attribute='url')

    class Meta:
        queryset = CrossRefDB.objects.all()
        allowed_methods = ['get']
        filtering = {'name': ALL,
                     'url': ALL}


class CrossRefResource(ModelResource):
    xrid = fields.CharField(attribute='xrid')
    crossrefdb = fields.ToOneField(CrossRefDBResource, 'crossrefdb')
    gene = fields.ToOneField(GeneResource, 'gene')
    db_url = fields.CharField(attribute='specific_url')

    class Meta:
        queryset = CrossRef.objects.all()
        allowed_methods = ['get']
        filtering = {
            'xrid': ALL,
            'crossrefdb': ALL,
            'gene': ALL_WITH_RELATIONS}
